using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;

namespace WpfConsole {
  class MainWindowViewModel : INotifyPropertyChanged {

    #region Start

    public MainWindowViewModel() {
    }

    public void RunDemos() {
      QueueDemos();
      ExecuteQueue();
    }

    #endregion Start

    #region Select Demos to Run

    /// <summary>
    /// This method populates a queue of demos. 
    /// </summary>
    /// <remarks>
    /// Comment out all demos except those you wish to run. You can run
    /// them all in rapid-fire sequence, but be aware that some produce 
    /// output asynchronously, so that the output can becomes somewhat 
    /// interlaced. However, all lines of output are prefixed with the
    /// name of the sample method with which they are associated.
    /// </remarks>
    private void QueueDemos() {
      IncludeMainDemo();
    }

    private void IncludeMainDemo() {
      MainDemo aMainDemo = new MainDemo(OutputWriter);
      _demos.Enqueue(() => aMainDemo.Concurrency_RemoteModifyLocalModify());
      _demos.Enqueue(() => aMainDemo.Concurrency_RemoteModifyLocalDelete());
      _demos.Enqueue(() => aMainDemo.Concurrency_RemoteDeleteLocalModify());
      _demos.Enqueue(() => aMainDemo.Concurrency_RemoteDeleteLocalDelete());
    }

    #endregion Select Demos to Run

    #region Data Bound Properties

    public string Output {
      get {
        return _output;
      }
      set {
        _output = value;
        RaisePropertyChanged("Output");
      }
    }

    #endregion Data Bound Properties

    #region INotifyPropertyChanged

    public event PropertyChangedEventHandler PropertyChanged = delegate { };

    protected void RaisePropertyChanged(string propertyName) {
      PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
    }

    #endregion

    #region Utility Methods

    public void OutputWriter(string output) {
      _outputStringBuilder.Append(output);
      this.Output = _outputStringBuilder.ToString();
      ExecuteQueue();
    }

    private void ExecuteQueue() {
      if (_demos.Count > 0) {
        var demoLaunchMethod = _demos.Dequeue();
        demoLaunchMethod();
      }
    }

    #endregion Utility Methods

    #region Private Fields

    private Queue<Action> _demos = new Queue<Action>();
    private StringBuilder _outputStringBuilder = new StringBuilder();
    private string _output = "";

    #endregion Private Fields

  }
}
