﻿Imports System.Text
Imports System.ComponentModel
Imports System.Collections.ObjectModel
Imports System.Linq.Expressions
Imports System.Collections
Imports System.Reflection

Imports IdeaBlade.EntityModel
Imports IdeaBlade.Linq

Namespace WpfConsole

  Public Class MainDemo
	Private _resultsReporter As ResultsReporter

	Public Sub New(ByVal callbackMethod As ResultsReporter)
	  Utility.InitializeDemo(_nameDemo, callbackMethod, _resultsReporter, _localOutput)
	End Sub

	#Region "Portals"

	Public Sub Concurrency_RemoteModifyLocalModify()
	  _localOutput.Append(String.Format("[{0}] Starting..." & vbLf, _nameConcurrency_RemoteModifyLocalModify))
	  ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteModifyLocalModify)
	End Sub

	Public Sub Concurrency_RemoteModifyLocalDelete()
	  _localOutput.Append(String.Format("[{0}] Starting..." & vbLf, _nameConcurrency_RemoteModifyLocalDelete))
	  ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteModifyLocalDelete)
	End Sub

	Public Sub Concurrency_RemoteDeleteLocalModify()
	  _localOutput.Append(String.Format("[{0}] Starting..." & vbLf, _nameConcurrency_RemoteDeleteLocalModify))
	  ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteDeleteLocalModify)
	End Sub

	Public Sub Concurrency_RemoteDeleteLocalDelete()
	  _localOutput.Append(String.Format("[{0}] Starting..." & vbLf, _nameConcurrency_RemoteDeleteLocalDelete))
	  ConcurrencyConflictSimulator(ConcurrencyScenario.RemoteDeleteLocalDelete)
	End Sub

	#End Region ' Portals

	#Region "ConcurrencyConflictSimulator"

	Private Sub ConcurrencyConflictSimulator(ByVal scenario As ConcurrencyScenario)
	  Dim nameOfExecutingTest As String = GetOutputLabelForExecutingTest(scenario)
	  _localOutput.Append(String.Format(vbLf & vbLf & "[{0}] Started..." & vbLf, nameOfExecutingTest))

	  Dim emLocal As New NorthwindIBModelManager() ' represents environment of local user
	  Dim emRemote As New NorthwindIBModelManager() ' represents environment of remote user

	  ' Create and save a new Customer and a related Order. 
	  ' We're doing this instead of retrieving existing data simply in order
	  ' to leave the latter untouched by this demo.
	  _theCustomerInstanceLocal = CreateAndSaveCustomerAndRelatedOrder(emLocal)

	  ' Get a second instance of the same Customer into a different EntityManager.
	  ' This simulates the scenario of a second User somewhere across the building 
	  ' or across the planet working with the same Customer record.
	  _theCustomerInstanceRemote = _theCustomerInstanceLocal.ToQuery().With(emRemote).FirstOrNullEntity()

	  ModifyInstances(scenario, _theCustomerInstanceLocal, _theCustomerInstanceRemote)

	  SavePendingChanges(nameOfExecutingTest, emRemote)
	  SavePendingChanges(nameOfExecutingTest, emLocal)

	  Utility.FlushOutputBuffer(_localOutput, _resultsReporter)
	End Sub

	Private Shared Function CreateAndSaveCustomerAndRelatedOrder(ByVal em As NorthwindIBModelManager) As Customer
	  Dim theCustomerInstance1 As New Customer()
	  theCustomerInstance1.CustomerID = Guid.NewGuid()
	  theCustomerInstance1.CompanyName = "theCustomerInstance1"
	  em.AddEntity(theCustomerInstance1)

	  Dim newOrder1 As New Order()
	  em.AddEntity(newOrder1)
	  theCustomerInstance1.Orders.Add(newOrder1)

	  em.SaveChanges()

	  Return theCustomerInstance1
	End Function

	''' <summary>
	''' Save the changed entities if possible, even if this requires
	''' overriding a concurrency conflict.
	''' </summary>
	''' <param name="nameOfExecutingTest"></param>
	''' <param name="em"></param>
	''' <remarks>
	''' Demonstrates techniques for overriding a concurrency conflict under
	''' different conflict scenarios. In a production app, you would not do
	''' that as a matter of course; on the other hand, you might wish to do it 
	''' under certain business scenarios. This shows you how to do it if you need to.
	''' </remarks>
	Private Sub SavePendingChanges(ByVal nameOfExecutingTest As String, ByVal em As NorthwindIBModelManager)
	  _localOutput.Append(String.Format("[{0}] Entities to be saved..." & vbLf, nameOfExecutingTest))
	  For Each anEntity As Entity In em.FindEntities(EntityState.AnyAddedModifiedOrDeleted)
		_localOutput.Append(String.Format(vbTab & "{0}" & vbLf, anEntity.EntityAspect.EntityKey.ToString()))
	  Next anEntity

	  Dim theSaveResult As SaveResult = Nothing
	  Dim entitySaved As Entity = Nothing
	  Dim saveSucceeded As Boolean = False
	  Do While Not saveSucceeded
		If _unsuccessfulSaveAttempts > _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS Then
		  Dim msg As String = "[{0}] Exceeded limit ({1}) on unsuccessful save attempts!" & vbLf & "This could occur if the number of conflicted entities submitted" & vbLf & "for saving simply exceeds that limit, or if the attempt to ready" & vbLf & "a conflicted entity for saving isn't working!" & vbLf
		  _localOutput.Append(String.Format(msg, nameOfExecutingTest, _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS))
		  Exit Do
		End If

		Try
		  ' ***************************************
		  theSaveResult = em.SaveChanges()
		  ' ***************************************
		  If theSaveResult.Ok Then
			saveSucceeded = True
			entitySaved = CType(theSaveResult.SavedEntities(0), Entity)
		  Else
			_unsuccessfulSaveAttempts += 1

			If theSaveResult.WasExceptionHandled Then
			  _localOutput.Append(String.Format("[{0}] Problem saving: {1}" & vbLf, nameOfExecutingTest, theSaveResult.HandledException.Message))
			Else
			  _localOutput.Append(String.Format("[{0}] Problem saving: {1}" & vbLf, nameOfExecutingTest, "Exception was not handled"))
			End If
		  End If
		Catch ex As Exception
		  Dim saveEx As EntityManagerSaveException = Nothing
		  If Not(TypeOf ex Is EntityManagerSaveException) Then
			_localOutput.Append(String.Format("[{0}] Error on save attempt: {1}" & vbLf, nameOfExecutingTest, ex.Message))
		  Else
			' Identify the concurrency scenario...
			saveEx = CType(ex, EntityManagerSaveException)
			Dim entityWithError As Entity = CType(saveEx.EntitiesWithErrors(0), Entity)
			Dim entityWithErrorType As Type = entityWithError.GetType()

			Dim msg As String = "[{0}] Error on attempt to save: {1}" & vbLf & vbTab & "{2}" & vbLf & vbLf & vbTab & "Applying fix to permit forced save..." & vbLf
			_localOutput.Append(String.Format(msg, nameOfExecutingTest, entityWithError.EntityAspect.EntityKey.ToString(), ex.Message))

			Dim emTemp As New NorthwindIBModelManager()
			Dim state As EntityState = entityWithError.EntityAspect.EntityState

			' Get a copy of the target entity from the database, if it still exists there...
			Dim remoteEntitiesQuery As EntityKeyQuery = entityWithError.EntityAspect.EntityKey.ToKeyQuery()
			Dim remoteEntities As List(Of Object) = remoteEntitiesQuery.With(emTemp).Execute().ToList()

			' The code in the following switch construct takes action to put the conflicted
			' entity in a state such that it will be saved successfully if submitted again (as
			' will happen in this save loop). As such, it causes the app to behave as if
			' concurrency checking were not even activated (Last One In Wins).
			'
			' In a production app, having bothered to check for concurrency conflicts in the first
			' place, you would only conditionally execute this logic to override the conflict. 
			' For example, you might do so after notifying the end user of the conflict and 
			' receiving an instruction from her to perform the save in spite of it; or you might 
			' have business logic that decides when to force the save (or when to permit the end 
			' user to decide).

			Select Case state
			  Case EntityState.Deleted
				' The "local" instance was marked for deletion.
				' Since we got a concurrency conflict, the remote entity was either
				' modified or deleted. If it exists, it was modified.

				If remoteEntities.Count = 0 Then
				  ' Discovered scenario: ConcurrencyScenario.RemoteDeleteLocalDelete
				  saveSucceeded = True
				  ' No further action is required. The local user wanted the record deleted,
				  ' and it has been deleted, albeit by the remote user. Since we know all of
				  ' this, we could inform the local user about exactly what happened, if
				  ' we deem that important to do.
				Else
				  ' Discovered scenario: ConcurrencyScenario.RemoteModifyLocalDelete
				  em.RefetchEntity(entityWithError, MergeStrategy.PreserveChangesUpdateOriginal)
				End If

			  Case EntityState.Modified
				' The "local" instance was modified.
				' Since we got a concurrency conflict, the remote entity was either
				' modified or deleted. If it exists, it was modified.

				If remoteEntities.Count = 0 Then
				  ' Discovered scenario: ConcurrencyScenario.RemoteDeleteLocalModify
				  entityWithError.EntityAspect.SetAdded()
				Else
				  ' Discovered scenario: ConcurrencyScenario.RemoteModifyLocalModify
				  em.RefetchEntity(entityWithError, MergeStrategy.PreserveChangesUpdateOriginal)
				End If

			  Case Else
				Throw New Exception("Unexpected EntityState: " & state.ToString())
			End Select
		  End If
		End Try
	  Loop
	  If saveSucceeded Then
		_localOutput.Append(String.Format("[{0}] Saved succeeded!" & vbLf, nameOfExecutingTest))
	  End If
	End Sub

	Private Shared Sub ModifyInstances(ByVal scenario As ConcurrencyScenario, ByVal theCustomerInstanceLocal As Customer, ByVal theCustomerInstanceRemote As Customer)
	  Select Case scenario
		Case ConcurrencyScenario.RemoteModifyLocalModify
		  ModifyCustomer(theCustomerInstanceRemote, "R")
		  ModifyCustomer(theCustomerInstanceLocal, "L")
		Case ConcurrencyScenario.RemoteModifyLocalDelete
		  ModifyCustomer(theCustomerInstanceRemote, "R")
		  MarkCustomerForDeletion(theCustomerInstanceLocal)
		Case ConcurrencyScenario.RemoteDeleteLocalModify
		  MarkCustomerForDeletion(theCustomerInstanceRemote)
		  ModifyCustomer(theCustomerInstanceLocal, "L")
		Case ConcurrencyScenario.RemoteDeleteLocalDelete
		  MarkCustomerForDeletion(theCustomerInstanceRemote)
		  MarkCustomerForDeletion(theCustomerInstanceLocal)
		Case Else
		  Throw New Exception("Unknown ConcurrencyScenario. Aborting...")
	  End Select
	End Sub

	Private Shared Sub ModifyCustomer(ByVal theCustomerInstance As Customer, ByVal suffix As String)
	  theCustomerInstance.CompanyName += suffix
	End Sub

	Private Shared Sub MarkCustomerForDeletion(ByVal theCustomerInstance As Customer)
	  Dim em As NorthwindIBModelManager = CType(theCustomerInstance.EntityAspect.EntityManager, NorthwindIBModelManager)

	  ' Related Orders must be deleted before we can delete this Customer...
	  Dim orders As List(Of Order) = theCustomerInstance.Orders.ToList()
	  For i As Integer = orders.Count - 1 To 0 Step -1
		orders(i).EntityAspect.Delete()
	  Next i

	  theCustomerInstance.EntityAspect.Delete()
	End Sub

	Private Shared Function GetOutputLabelForExecutingTest(ByVal scenario As ConcurrencyScenario) As String
	  Dim nameOfExecutingTest As String
	  Select Case scenario
		Case ConcurrencyScenario.RemoteModifyLocalModify
		  nameOfExecutingTest = "RemoteModify/LocalModify"
		Case ConcurrencyScenario.RemoteModifyLocalDelete
		  nameOfExecutingTest = "RemoteModify/LocalDelete"
		Case ConcurrencyScenario.RemoteDeleteLocalModify
		  nameOfExecutingTest = "RemoteDelete/LocalModify"
		Case ConcurrencyScenario.RemoteDeleteLocalDelete
		  nameOfExecutingTest = "RemoteDelete/LocalDelete"
		Case Else
		  nameOfExecutingTest = ""
		  Throw New Exception("Unknown ConcurrencyScenario. Aborting...")
	  End Select
	  Return nameOfExecutingTest
	End Function

	#End Region ' ConcurrencyConflictSimulator

	#Region "Private Fields"

	Private Const _nameDemo As String = "MainDemo (MDM)"
	Private Const _nameConcurrency_RemoteModifyLocalModify As String = "MDM:Concurrency_RemoteModifyLocalModify"
	Private Const _nameConcurrency_RemoteDeleteLocalModify As String = "MDM:Concurrency_RemoteDeleteLocalModify"
	Private Const _nameConcurrency_RemoteModifyLocalDelete As String = "MDM:Concurrency_RemoteModifyLocalDelete"
	Private Const _nameConcurrency_RemoteDeleteLocalDelete As String = "MDM:Concurrency_RemoteDeleteLocalDelete"

	Private _localOutput As StringBuilder

	Private _theCustomerInstanceLocal As Customer
	Private _theCustomerInstanceRemote As Customer

	Private Enum ConcurrencyHandlerStrategy
	  SetAdded
	  Refetch
	End Enum

	Private Enum ConcurrencyScenario
	  RemoteModifyLocalModify
	  RemoteModifyLocalDelete
	  RemoteDeleteLocalModify
	  RemoteDeleteLocalDelete
	End Enum

	Private _unsuccessfulSaveAttempts As Integer = 0
	Private Const _MAX_UNSUCCESSFUL_SAVE_ATTEMPTS As Integer = 10

	#End Region ' Private Fields

  End Class
End Namespace
