﻿Imports System.Text
Imports System.ComponentModel
Imports System.Collections.ObjectModel
Imports System.Linq.Expressions
Imports IdeaBlade.EntityModel
Imports System.Collections
Imports IdeaBlade.Linq
Imports System.Reflection

Namespace WpfConsole
  Friend Class MainWindowViewModel
    Implements INotifyPropertyChanged

#Region "Start"

    Public Sub New()
    End Sub

    Public Sub RunDemos()
      QueueDemos()
      ExecuteQueue()
    End Sub

#End Region ' Start

#Region "Select Demos to Run"

    ''' <summary>
    ''' This method populates a queue of demos. 
    ''' </summary>
    ''' <remarks>
    ''' Comment out all demos except those you wish to run. You can run
    ''' them all in rapid-fire sequence, but be aware that some produce 
    ''' output asynchronously, so that the output can becomes somewhat 
    ''' interlaced. However, all lines of output are prefixed with the
    ''' name of the sample method with which they are associated.
    ''' </remarks>
    Private Sub QueueDemos()
      IncludeMainDemo()
    End Sub

    Private Sub IncludeMainDemo()
      Dim aMainDemo As New MainDemo(AddressOf OutputWriter)
      _demos.Enqueue(Sub() aMainDemo.Concurrency_RemoteModifyLocalModify())
      _demos.Enqueue(Sub() aMainDemo.Concurrency_RemoteModifyLocalDelete())
      _demos.Enqueue(Sub() aMainDemo.Concurrency_RemoteDeleteLocalModify())
      _demos.Enqueue(Sub() aMainDemo.Concurrency_RemoteDeleteLocalDelete())
    End Sub

#End Region ' Select Demos to Run

#Region "Data Bound Properties"

    Public Property Output() As String
      Get
        Return _output
      End Get
      Set(ByVal value As String)
        _output = value
        RaisePropertyChanged("Output")
      End Set
    End Property


#End Region ' Data Bound Properties

#Region "INotifyPropertyChanged"

    Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged

    Protected Sub RaisePropertyChanged(ByVal propertyName As String)
      RaiseEvent PropertyChanged(Me, New PropertyChangedEventArgs(propertyName))
    End Sub

#End Region

#Region "Utility Methods"

    Public Sub OutputWriter(ByVal output As String)
      _outputStringBuilder.Append(output)
      Me.Output = _outputStringBuilder.ToString()
      ExecuteQueue()
    End Sub

    Private Sub ExecuteQueue()
      If _demos.Count > 0 Then
        Dim demoLaunchMethod = _demos.Dequeue()
        demoLaunchMethod()
      End If
    End Sub

#End Region ' Utility Methods

#Region "Private Fields"

    Private _demos As New Queue(Of Action)()
    Private _outputStringBuilder As New StringBuilder()
    Private _output As String = ""

#End Region ' Private Fields

  End Class
End Namespace
